import logging
import threading
import time

import requests

from .server import JsonGet, http_server

logging.basicConfig()

PORT = 9999
SLEEP_TIME = 10


class RequestThenStopThread(threading.Thread):
    def __init__(self, server):
        threading.Thread.__init__(self)
        self.response = None
        self.server = server

    def run(self):
        time.sleep(SLEEP_TIME)
        result = requests.get(f"http://localhost:{PORT}/test")
        result.raise_for_status()
        self.response = result.json()
        self.server.stop()


class TestServer:

    def teardown_method(self, method) -> None:
        if self.server is not None:
            self.server.stop()

    def test_should_block_main_thread(self):
        # given
        time_before_server_start = time.time()

        self.server = http_server(PORT, [JsonGet("/test", lambda params: {'status': 'ok'})], thread_count=2)

        request_thread = RequestThenStopThread(self.server)
        request_thread.start()

        self.server.start(block_caller_thread=True)

        # when
        time_after_server_start = time.time()

        # then
        assert SLEEP_TIME < time_after_server_start - time_before_server_start
        assert 'ok' == request_thread.response['status']



import json
import logging
import mimetypes
import os
import socket
import threading
import urllib.parse as urlparse
from http.server import BaseHTTPRequestHandler, HTTPServer
from io import BytesIO

import multipart


class Response:
    def code(self):
        return 200

    def headers(self) -> dict:
        return {}

    def response_body_bytes(self):
        return "".encode("utf-8")


class ErrorResponse(Response):
    def __init__(self, code, message):
        self._code = code
        self._message = message

    def code(self):
        return self._code

    def headers(self) -> dict:
        return {'Content-type': 'application/json'}

    def response_body_bytes(self):
        return json.dumps({'error': self._message}, indent=4, default=lambda o: o.__dict__).encode('utf-8')


class RedirectResponse(Response):
    def __init__(self, location):
        self._location = location

    def code(self):
        return 301

    def headers(self) -> dict:
        return {'Location': self._location}


class JsonOkResponse(Response):
    def __init__(self, payload):
        self._payload = payload

    def headers(self) -> dict:
        return {'Content-type': 'application/json'}

    def response_body_bytes(self):
        return json.dumps(self._payload, indent=4, default=lambda o: o.__dict__).encode('utf-8')


class StaticResourceResponse(Response):
    def __init__(self, mime, data):
        self._mime = mime
        self._data = data

    def headers(self) -> dict:
        return {'Content-type': self._mime}

    def response_body_bytes(self):
        return self._data


class Action:
    def can_handle(self, method, path, params, payload):
        return False

    def handle(self, method, path, params, payload) -> Response:
        return None


class JsonGet(Action):
    def __init__(self, endpoint, callable):
        self.path = endpoint
        self.callable = callable

    def can_handle(self, method, path, params, payload):
        return method == 'GET' and path == self.path

    def handle(self, method, path, params, payload) -> Response:
        output = self.callable(params)
        return JsonOkResponse(output)


class JsonPost(Action):
    def __init__(self, endpoint, callable):
        self.path = endpoint
        self.callable = callable

    def can_handle(self, method, path, params, payload):
        return method == 'POST' and path == self.path

    def handle(self, method, path, params, payload) -> Response:
        if len(payload) > 0:
            try:
                payload = json.loads(payload)
            except Exception as e:
                raise ValidationException(e)
        output = self.callable(params, payload)
        return JsonOkResponse(output)


class ProcessFileUploadThenRedirect(Action):
    def __init__(self, endpoint, callable):
        self.path = endpoint
        self.callable = callable

    def can_handle(self, method, path, params, payload):
        return method == 'POST' and path == self.path

    def handle(self, method, path, params, payload) -> Response:
        extractor = FileUploadExtractor(payload)
        if not extractor.is_file_uploaded():
            raise ValidationException("No uploaded file found")
        location = self.callable(params, extractor.extract_uploaded_files())
        return RedirectResponse(location)


class Redirect(Action):
    def __init__(self, path_from, path_to):
        self.path_from = path_from
        self.path_to = path_to

    def can_handle(self, method, path, params, payload):
        return path == self.path_from

    def handle(self, method, path, params, payload) -> Response:
        return RedirectResponse(self.path_to)


class StaticResources(Action):
    def __init__(self, url_path_prefix, dir):
        self._url_prefix = url_path_prefix
        self._dir = dir
        if not self._url_prefix.endswith('/'):
            self._url_prefix += '/'

    def can_handle(self, method, path, params, payload):
        return method == 'GET' and path.startswith(self._url_prefix) and '/' not in path[len(self._url_prefix):]

    def handle(self, method, path, params, payload) -> Response:
        path = path[len(self._url_prefix):]
        optional_slash = '/' if not self._dir.endswith('/') and not path.startswith('/') else ''
        file = self._dir + optional_slash + path
        if os.path.isfile(file):
            mime, encoding = mimetypes.guess_type(file)
            with open(file, 'rb') as f:
                data = f.read()
            return StaticResourceResponse(mime, data)
        else:
            logging.getLogger('web').info("Static resource %s not found" % file)
            return ErrorResponse(404, "File not found: %s" % path)


class ValidationException(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class ServerController:
    def __init__(self, threads: list, sock):
        self.threads = threads
        self.sock = sock
        self.semaphore = threading.Semaphore(0)

    def start(self, block_caller_thread: bool = False):
        for thread in self.threads:
            thread.start()
        if block_caller_thread:
            self.semaphore.acquire()

    def stop(self):
        for thread in self.threads:
            thread.prepare_shutdown()
        self.sock.close()
        for thread in self.threads:
            thread.shutdown()
        self.semaphore.release()


class UploadedFile:
    def __init__(self, part: multipart.MultipartPart):
        self.form_name = part.name
        self.original_file_name = part.filename
        self.__part = part

    def save_as(self, path):
        self.__part.save_as(path)


class FileUploadExtractor:
    def __init__(self, payload):
        lines = payload.splitlines()
        if len(lines) > 2 and lines[0].startswith(b'--') and b'Content-Disposition: form-data' in lines[1]:
            self.payload = payload
            self.boundary = lines[0][2:]
        else:
            self.payload = None
            self.boundary = None

    def is_file_uploaded(self):
        return self.payload is not None

    def extract_uploaded_files(self):
        parts = multipart.MultipartParser(BytesIO(self.payload), self.boundary).parts()
        result = {}
        for part in parts:
            result[part.name] = UploadedFile(part)
        return result


def http_server(port: int, actions: list, thread_count: int = 10) -> ServerController:
    logger = logging.getLogger('web')

    logger.info('Starting httpserver on port %s in %s threads' % (port, thread_count))

    socket_address = ('', port)
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.bind(socket_address)
    sock.listen(5)

    class ServerThread(threading.Thread):
        def __init__(self, i):
            threading.Thread.__init__(self)
            self.i = i
            self.daemon = False
            self.httpd = None
            self.shutting_down = False

        def run(self):
            self.httpd = HTTPServer(socket_address, RequestHandler, False)

            # Prevent the HTTP server from re-binding every handler.
            # https://stackoverflow.com/questions/46210672/
            self.httpd.socket = sock
            self.httpd.server_bind = self.server_close = lambda self: None

            try:
                self.httpd.serve_forever()
            except Exception as e:
                if not self.shutting_down:
                    raise e

        def prepare_shutdown(self):
            self.shutting_down = True

        def shutdown(self):
            if self.httpd is not None:
                self.httpd.shutdown()

    class RequestHandler(BaseHTTPRequestHandler):
        def do_GET(self):
            logger.debug("Accepted HTTP request %s on thread %s" % (self.path, threading.current_thread().name))
            try:
                path, params = self.__parse(self.path)
                for action in actions:
                    if action.can_handle('GET', path, params, {}):
                        output = action.handle('GET', path, params, {})
                        return self.__respond(output)
                return self.__respond(ErrorResponse(404, "File not found"))
            except ValidationException as e:
                self.__respond(ErrorResponse(400, "URL parse exception: %s" % str(e)))
                logger.error("ValidationException", e)
            except Exception as e:
                self.__respond(ErrorResponse(500, str(e)))
                logger.error("Exception", e)

        def do_POST(self):
            logger.debug("Accepted HTTP request %s on thread %s" % (self.path, threading.current_thread().name))
            try:
                path, params = self.__parse(self.path)
                content_length = int(self.headers['Content-Length']) if 'Content-Length' in self.headers else 0
                payload = self.rfile.read(content_length) if content_length > 0 else ""
                for action in actions:
                    if action.can_handle('POST', path, params, payload):
                        output = action.handle('POST', path, params, payload)
                        return self.__respond(output)
                return self.__respond(ErrorResponse(404, "File not found"))
            except ValidationException as e:
                self.__respond(ErrorResponse(400, "URL parse exception: %s" % str(e)))
                logger.error("ValidationException", e)
            except Exception as e:
                self.__respond(ErrorResponse(500, str(e)))
                logger.error("Exception", e)

        def __respond(self, output):
            code = output.code()
            headers = output.headers()
            payload = output.response_body_bytes()
            logger.debug("Returning HTTP %s %s with payload of %s bytes" % (code, headers, len(payload)))
            self.send_response(code)
            for header in headers:
                self.send_header(header, headers[header])
            self.end_headers()
            self.wfile.write(payload)

        @staticmethod
        def __parse(url):
            try:
                url_parsed = urlparse.urlparse(url)
                url_params = urlparse.parse_qs(url_parsed.query)
                url_path = url_parsed.path
                return url_path, url_params
            except Exception as e:
                raise ValidationException(e)

    threads = [ServerThread(i) for i in range(thread_count)]

    return ServerController(threads, sock)



from src.actions_server import *

ACTIONS = [
    JsonGet("/get", lambda params: {"response": "ok from GET action"}),
    JsonPost("/post", lambda params, body: {"response": "ok from POST action"}),
    Redirect("/", "/get"),
    StaticResources("/static", "./src/web")
]

server = http_server(port=80, actions=ACTIONS, thread_count=5)
try:
    server.start(block_caller_thread=True)
finally:
    server.stop()



from .server import *

__all__ = [
    'http_server',
    'Response',
    'Action',
    'JsonGet',
    'JsonPost',
    'Redirect',
    'StaticResources',
    'ServerController',
    'ProcessFileUploadThenRedirect',
    'ValidationException',
    'UploadedFile',
    'JsonOkResponse',
    'ErrorResponse'
]



import logging
import pytest

import requests

from .server import JsonGet, http_server, Action, JsonPost, Redirect, StaticResources

logging.basicConfig()

PORT = 9999


class TestServer:

    def teardown_method(self, method) -> None:
        if self.server is not None:
            self.server.stop()

    def test_should_call_get_action(self):
        # given
        self._start_http_server(JsonGet("/test", lambda params: {'status': 'ok'}))

        # when
        result = requests.get(f"http://localhost:{PORT}/test")

        # then
        assert 200 == result.status_code
        assert 'ok' == result.json()['status']

    def test_should_call_get_action_with_params(self):
        # given
        self._start_http_server(JsonGet("/test", lambda params: {'params': params}))

        # when
        result = requests.get(f"http://localhost:{PORT}/test?a=0&b=6&b=7")

        # then
        assert 200 == result.status_code
        assert ['0'] == result.json()['params']['a']
        assert ['6', '7'] == result.json()['params']['b']

    def test_should_call_post_action(self):
        # given
        self._start_http_server(JsonPost("/test", lambda params, body: {'status': 'ok'}))

        # when
        result = requests.post(f"http://localhost:{PORT}/test")

        # then
        assert 200 == result.status_code
        assert 'ok' == result.json()['status']

    def test_should_call_post_action_with_params_and_body(self):
        # given
        self._start_http_server(JsonPost("/test", lambda params, body: {'params': params, 'body': body}))

        # when
        result = requests.post(f"http://localhost:{PORT}/test?a=0&b=6&b=7", data='{"hello": "POST BODY"}')

        # then
        assert 200 == result.status_code
        assert ['0'] == result.json()['params']['a']
        assert ['6', '7'] == result.json()['params']['b']
        assert 'POST BODY' == result.json()['body']['hello']

    def test_should_call_post_action_without_body(self):
        # given
        self._start_http_server(JsonPost("/test", lambda params, body: {'body': body}))

        # when
        result = requests.post(f"http://localhost:{PORT}/test?a=0&b=6&b=7")

        # then
        assert 200 == result.status_code
        assert '' == result.json()['body']

    def test_should_return_4xx_when_post_action_with_unparsable_body(self):
        # given
        self._start_http_server(JsonPost("/test", lambda params, body: {'body': body}))

        # when
        result = requests.post(f"http://localhost:{PORT}/test", data='non-json string')

        # then
        assert 400 == result.status_code

    def test_should_redirect(self):
        # given
        self._start_http_server(Redirect("/test", "http://example.com"))

        # when:
        result = requests.get(f'http://localhost:{PORT}/test', allow_redirects=False)

        # then
        assert 301 == result.status_code
        assert 'http://example.com' == result.headers['Location']

    @pytest.mark.parametrize("filename, expected_length, expected_content_type", [
        ("text.txt", 9, "text/plain"),
        ("image.png", 81618, "image/png"),
        ("document.pdf", 38078, "application/pdf")
    ])
    def test_should_serve_static_resources(self, filename, expected_length, expected_content_type):
        # given
        self._start_http_server(StaticResources("/static", "./test-resources"))

        # when:
        result = requests.get(f'http://localhost:{PORT}/static/{filename}')

        # then
        assert 200 == result.status_code
        assert expected_length == len(result.content)
        assert expected_content_type == result.headers['Content-Type']

    def test_should_serve_static_resources_regardless_slash(self):
        # given
        self._start_http_server(StaticResources("/static/", "./test-resources/"))

        # when:
        result = requests.get(f'http://localhost:{PORT}/static/text.txt')

        # then
        assert 200 == result.status_code
        assert 9 == len(result.content)

    def test_should_return_404_if_static_resources_not_found(self):
        # given
        self._start_http_server(StaticResources("/static/", "./test-resources/"))

        # when:
        result = requests.get(f'http://localhost:{PORT}/static/nonexisting.txt')

        # then
        assert 404 == result.status_code

    def test_should_return_404_if_action_not_found(self):
        # given
        self._start_http_server(JsonGet("/test", lambda params: {'status': 'ok'}))

        # when:
        result = requests.get(f'http://localhost:{PORT}/nonexisting-endpoint')

        # then
        assert 404 == result.status_code

    def test_should_find_action_in_order_of_declaration(self):
        # given
        self._start_http_server([
            JsonGet("/test", lambda params: {'status': '1'}),
            JsonGet("/test", lambda params: {'status': '2'}),
        ])

        # when:
        result = requests.get(f'http://localhost:{PORT}/test')

        # then
        assert 200 == result.status_code
        assert '1' == result.json()['status']

    def _start_http_server(self, actions):
        actions = actions if isinstance(actions, list) else [actions]
        self.server = http_server(PORT, actions, thread_count=1)
        self.server.start(block_caller_thread=False)



import os

import setuptools
import time

with open("README.md", "r") as fh:
    long_description = fh.read()

postfix = "" if os.getenv("RELEASE", "0") == "1" else ".dev%s" % round(time.time())

setuptools.setup(
    name="actions-server",
    version="0.0.9%s" % postfix,
    description="A very simple, multi-threaded HTTP server",
    author="Rafał Zarajczyk",
    author_email="rzarajczyk@gmail.com",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/rzarajczyk/actions-server",
    keywords=["HTTP", "SERVER"],
    packages=['actions_server'],
    package_dir={'actions_server': './src/actions_server'},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=["requests>=2.27.1", "multipart>=0.2.4"],
)
